@@ -18,10 +18,7 @@ use arkworks_gadgets::{
 		Private as LeafPrivateInputs, Public as LeafPublicInputs,
 	},
 	merkle_tree::{constraints::PathVar, Config as MerkleConfig, Path},
-	set::membership::{
-		constraints::{PrivateVar as SetPrivateInputsVar, SetMembershipGadget},
-		Private as SetPrivateInputs,
-	},
+	set::constraints::SetGadget,
 };
 
 pub struct VAnchorCircuit<
@@ -43,8 +40,7 @@ pub struct VAnchorCircuit<
 
 	leaf_private_inputs: Vec<LeafPrivateInputs<F>>, // amount, blinding
 	keypair_inputs: Vec<Keypair<F, H>>,
-	leaf_public_input: LeafPublicInputs<F>,          // chain_id
-	set_private_inputs: Vec<SetPrivateInputs<F, M>>, // diffs
+	leaf_public_input: LeafPublicInputs<F>, // chain_id
 	root_set: [F; M],
 	hasher_params_w2: H::Parameters,
 	hasher_params_w4: H::Parameters,
@@ -92,7 +88,6 @@ where
 		leaf_private_inputs: Vec<LeafPrivateInputs<F>>,
 		keypair_inputs: Vec<Keypair<F, H>>,
 		leaf_public_input: LeafPublicInputs<F>,
-		set_private_inputs: Vec<SetPrivateInputs<F, M>>,
 		root_set: [F; M],
 		hasher_params_w2: H::Parameters,
 		hasher_params_w4: H::Parameters,
@@ -111,7 +106,6 @@ where
 			leaf_private_inputs,
 			keypair_inputs,
 			leaf_public_input,
-			set_private_inputs,
 			root_set,
 			hasher_params_w2,
 			hasher_params_w4,
@@ -142,8 +136,7 @@ where
 		in_path_indices_var: &[FpVar<F>],
 		in_path_elements_var: &[PathVar<F, C, HGT, LHGT, K>],
 		in_nullifier_var: &[HG::OutputVar],
-		root_set_var: &[FpVar<F>],
-		set_input_private_var: &[SetPrivateInputsVar<F, M>],
+		set_gadget: &SetGadget<F>,
 	) -> Result<FpVar<F>, SynthesisError> {
 		let mut sums_ins_var = FpVar::<F>::zero();
 
@@ -177,12 +170,9 @@ where
 			// Add the roots and diffs signals to the vanchor circuit
 			let roothash = &in_path_elements_var[tx].root_hash(&in_utxo_hasher_var)?;
 			let in_amount_tx = &leaf_private_var[tx].amount;
-			let check = SetMembershipGadget::check_is_enabled(
-				&roothash,
-				&root_set_var.to_vec(),
-				&set_input_private_var[tx],
-				in_amount_tx,
-			)?;
+
+			// Check membership if in_amount is non zero
+			let check = set_gadget.check_membership_enabled(&roothash, in_amount_tx)?;
 			check.enforce_equal(&Boolean::TRUE)?;
 
 			sums_ins_var += in_amount_tx;
@@ -269,7 +259,6 @@ where
 		let ext_data_hash = self.ext_data_hash.clone();
 		let leaf_private_inputs = self.leaf_private_inputs.clone();
 		let leaf_public_input = self.leaf_public_input.clone();
-		let set_private_inputs = self.set_private_inputs.clone();
 		let root_set = self.root_set;
 		let hasher_params_w2 = self.hasher_params_w2.clone();
 		let hasher_params_w4 = self.hasher_params_w4.clone();
@@ -289,7 +278,6 @@ where
 			leaf_private_inputs,
 			keypair_inputs,
 			leaf_public_input,
-			set_private_inputs,
 			root_set,
 			hasher_params_w2,
 			hasher_params_w4,
@@ -331,7 +319,6 @@ where
 		let leaf_private = self.leaf_private_inputs; // amount, blinding
 		let keypair_inputs = self.keypair_inputs;
 		let leaf_public_input = self.leaf_public_input; // chain id
-		let set_private = self.set_private_inputs;
 		let root_set = self.root_set;
 		let hasher_params_w2 = self.hasher_params_w2;
 		let hasher_params_w4 = self.hasher_params_w4;
@@ -376,8 +363,6 @@ where
 			Vec::<LeafPrivateInputsVar<F>>::new_witness(cs.clone(), || Ok(leaf_private))?;
 		let inkeypair_var =
 			Vec::<KeypairVar<F, H, HG>>::new_witness(cs.clone(), || Ok(keypair_inputs))?;
-		let set_input_private_var =
-			Vec::<SetPrivateInputsVar<F, M>>::new_witness(cs.clone(), || Ok(set_private))?;
 		let in_path_elements_var =
 			Vec::<PathVar<F, C, HGT, LHGT, K>>::new_witness(cs.clone(), || Ok(paths))?;
 		let in_path_indices_var = Vec::<FpVar<F>>::new_witness(cs.clone(), || Ok(indices))?;
@@ -389,6 +374,7 @@ where
 			Vec::<LeafPublicInputsVar<F>>::new_witness(cs.clone(), || Ok(out_leaf_public))?;
 		let out_pubkey_var = Vec::<FpVar<F>>::new_witness(cs, || Ok(out_pubkey))?;
 
+		let set_gadget = SetGadget::new(root_set_var);
 		// verify correctness of transaction inputs
 		let sum_ins_var = Self::verify_input_var(
 			&hasher_params_w2_var,
@@ -400,8 +386,7 @@ where
 			&in_path_indices_var,
 			&in_path_elements_var,
 			&in_nullifier_var,
-			&root_set_var,
-			&set_input_private_var,
+			&set_gadget,
 		)?;
 
 		// verify correctness of transaction outputs
@@ -429,6 +414,8 @@ where
 
 #[cfg(test)]
 mod test {
+	use ark_std::vec;
+
 	use crate::{
 		ark_std::{One, Zero},
 		setup::{common::*, vanchor::VAnchorProverBn2542x2},
@@ -449,6 +436,76 @@ mod test {
 	use ark_snark::SNARK;
 	use ark_std::test_rng;
 
+	#[test]
+	fn should_create_proof_for_random_circuit() {
+		let rng = &mut test_rng();
+		let curve = Curve::Bn254;
+		let params2 = setup_params_x5_2::<BnFr>(curve);
+		let params3 = setup_params_x5_3::<BnFr>(curve);
+		let params4 = setup_params_x5_4::<BnFr>(curve);
+		let params5 = setup_params_x5_5::<BnFr>(curve);
+
+		// Set up a random circuit and make pk/vk pair
+		let prover = VAnchorProverBn2542x2::new(params2, params3, params4, params5);
+		let random_circuit = prover.clone().setup_random_circuit(rng).unwrap();
+		let (proving_key, verifying_key) = setup_keys::<Bn254, _, _>(random_circuit, rng).unwrap();
+
+		// Make a proof now
+		let public_amount = BnFr::from(10u32);
+		let ext_data_hash = BnFr::rand(rng);
+
+		// Input Utxos
+		let in_chain_id = BnFr::from(0u32);
+		let in_amount = BnFr::from(5u32);
+		let index = BnFr::from(0u32);
+		let in_utxo1 = prover
+			.new_utxo(in_chain_id, in_amount, Some(index), None, None, rng)
+			.unwrap();
+		let in_utxo2 = prover
+			.new_utxo(in_chain_id, in_amount, Some(index), None, None, rng)
+			.unwrap();
+		let in_utxos = [in_utxo1, in_utxo2];
+
+		// Output Utxos
+		let out_chain_id = BnFr::from(0u32);
+		let out_amount = BnFr::from(10u32);
+		let out_utxo1 = prover
+			.new_utxo(out_chain_id, out_amount, None, None, None, rng)
+			.unwrap();
+		let out_utxo2 = prover
+			.new_utxo(out_chain_id, out_amount, None, None, None, rng)
+			.unwrap();
+		let out_utxos = [out_utxo1, out_utxo2];
+
+		let leaf0 = in_utxo1.commitment;
+		let (in_path0, _) = prover.setup_tree(&vec![leaf0], 0).unwrap();
+		let root0 = in_path0.root_hash(&leaf0).unwrap().inner();
+		let leaf1 = in_utxo2.commitment;
+		let (in_path1, _) = prover.setup_tree(&vec![leaf1], 0).unwrap();
+		let root1 = in_path1.root_hash(&leaf1).unwrap().inner();
+
+		let in_leaves = [vec![leaf0], vec![leaf1]];
+		let in_indices = [0; 2];
+		let in_root_set = [root0, root1];
+
+		let (circuit, .., pub_ins) = prover
+			.setup_circuit_with_utxos(
+				public_amount,
+				ext_data_hash,
+				in_root_set,
+				in_indices,
+				in_leaves,
+				in_utxos,
+				out_utxos,
+			)
+			.unwrap();
+
+		let proof = prove::<Bn254, _, _>(circuit, &proving_key, rng).unwrap();
+		let res = verify::<Bn254>(&pub_ins, &verifying_key, &proof).unwrap();
+
+		assert!(res);
+	}
+
 	#[test]
 	fn should_create_circuit_and_prove_groth16_2_input_2_output() {
 		let rng = &mut test_rng();
@@ -459,29 +516,52 @@ mod test {
 		let params5: PoseidonParameters<BnFr> = setup_params_x5_5(curve);
 		let prover = VAnchorProverBn2542x2::new(params2, params3, params4, params5);
 
-		let public_amount = 6;
-		let recipient = vec![1u8; 20];
-		let relayer = vec![2u8; 20];
-		let ext_amount = 10;
-		let fee = 0;
+		let public_amount = BnFr::from(10u32);
+		let ext_data_hash = BnFr::rand(rng);
 
-		let in_chain_id = 0;
-		let in_amounts = [2; 2];
-		let out_chain_ids = [0; 2];
-		let out_amounts = [5; 2];
+		// Input Utxos
+		let in_chain_id = BnFr::from(0u32);
+		let in_amount = BnFr::from(5u32);
+		let index = BnFr::from(0u32);
+		let in_utxo1 = prover
+			.new_utxo(in_chain_id, in_amount, Some(index), None, None, rng)
+			.unwrap();
+		let in_utxo2 = prover
+			.new_utxo(in_chain_id, in_amount, Some(index), None, None, rng)
+			.unwrap();
+		let in_utxos = [in_utxo1, in_utxo2];
 
-		let (circuit, pub_ins, ..) = prover
-			.setup_circuit_with_data(
+		// Output Utxos
+		let out_chain_id = BnFr::from(0u32);
+		let out_amount = BnFr::from(10u32);
+		let out_utxo1 = prover
+			.new_utxo(out_chain_id, out_amount, None, None, None, rng)
+			.unwrap();
+		let out_utxo2 = prover
+			.new_utxo(out_chain_id, out_amount, None, None, None, rng)
+			.unwrap();
+		let out_utxos = [out_utxo1, out_utxo2];
+
+		let leaf0 = in_utxo1.commitment;
+		let (in_path0, _) = prover.setup_tree(&vec![leaf0], 0).unwrap();
+		let root0 = in_path0.root_hash(&leaf0).unwrap().inner();
+		let leaf1 = in_utxo2.commitment;
+		let (in_path1, _) = prover.setup_tree(&vec![leaf1], 0).unwrap();
+		let root1 = in_path1.root_hash(&leaf1).unwrap().inner();
+
+		let in_leaves = [vec![leaf0], vec![leaf1]];
+		let in_indices = [0; 2];
+		let in_root_set = [root0, root1];
+
+		let (circuit, .., pub_ins) = prover
+			.setup_circuit_with_utxos(
 				public_amount,
-				recipient,
-				relayer,
-				ext_amount,
-				fee,
-				in_chain_id,
-				in_amounts,
-				out_chain_ids,
-				out_amounts,
-				rng,
+				ext_data_hash,
+				in_root_set,
+				in_indices,
+				in_leaves,
+				in_utxos,
+				out_utxos,
 			)
 			.unwrap();
 
@@ -504,44 +584,47 @@ mod test {
 
 		let public_amount = BnFr::from(10u32);
 		let ext_data_hash = BnFr::rand(rng);
-		let arbitrary_data = VAnchorProverBn2542x2::setup_arbitrary_data(ext_data_hash);
 
-		let in_chain_ids = [0; 2];
-		let in_amounts = [5; 2];
-		let in_utxos = prover.new_utxos(in_chain_ids, in_amounts, rng).unwrap();
+		// Input Utxos
+		let in_chain_id = BnFr::from(0u32);
+		let in_amount = BnFr::from(5u32);
+		let index = BnFr::from(0u32);
+		let in_utxo1 = prover
+			.new_utxo(in_chain_id, in_amount, Some(index), None, None, rng)
+			.unwrap();
+		let in_utxo2 = prover
+			.new_utxo(in_chain_id, in_amount, Some(index), None, None, rng)
+			.unwrap();
+		let in_utxos = [in_utxo1, in_utxo2];
 
-		let out_chain_ids = [0; 2];
-		let out_amounts = [5; 2];
-		let out_utxos = prover.new_utxos(out_chain_ids, out_amounts, rng).unwrap();
+		// Output Utxos
+		let out_chain_id = BnFr::from(0u32);
+		let out_amount = BnFr::from(10u32);
+		let out_utxo1 = prover
+			.new_utxo(out_chain_id, out_amount, None, None, None, rng)
+			.unwrap();
+		let out_utxo2 = prover
+			.new_utxo(out_chain_id, out_amount, None, None, None, rng)
+			.unwrap();
+		let out_utxos = [out_utxo1, out_utxo2];
 
-		let commitments = in_utxos.iter().map(|x| x.commitment).collect::<Vec<BnFr>>();
-		let (in_indices, in_paths, in_set_private_inputs, _) =
-			prover.setup_tree_and_set(&commitments).unwrap();
+		let leaf0 = in_utxos[0].commitment;
+		let leaf1 = in_utxos[1].commitment;
+
+		let in_leaves = [vec![leaf0], vec![leaf1]];
+		let in_indices = [0; 2];
 
 		// Invalid root set
 		let in_root_set = [BnFr::rand(rng); 2];
 
-		let pub_ins = VAnchorProverBn2542x2::construct_public_inputs(
-			in_utxos[0].leaf_public.chain_id,
-			public_amount,
-			in_root_set.to_vec(),
-			in_utxos.iter().map(|x| x.nullifier).collect::<Vec<BnFr>>(),
-			out_utxos
-				.iter()
-				.map(|x| x.commitment)
-				.collect::<Vec<BnFr>>(),
-			ext_data_hash,
-		);
-
-		let circuit = prover
-			.setup_circuit(
+		let (circuit, .., pub_ins) = prover
+			.setup_circuit_with_utxos(
 				public_amount,
-				arbitrary_data,
-				in_utxos,
-				in_indices,
-				in_paths,
-				in_set_private_inputs,
+				ext_data_hash,
 				in_root_set,
+				in_indices,
+				in_leaves,
+				in_utxos,
 				out_utxos,
 			)
 			.unwrap();
@@ -565,44 +648,53 @@ mod test {
 
 		let public_amount = BnFr::from(10u32);
 		let ext_data_hash = BnFr::rand(rng);
-		let arbitrary_data = VAnchorProverBn2542x2::setup_arbitrary_data(ext_data_hash);
 
-		let in_chain_ids = [0; 2];
-		let in_amounts = [5; 2];
-		let mut in_utxos = prover.new_utxos(in_chain_ids, in_amounts, rng).unwrap();
+		// Input Utxos
+		let in_chain_id = BnFr::from(0u32);
+		let in_amount = BnFr::from(5u32);
+		let index = BnFr::from(0u32);
+		let mut in_utxo1 = prover
+			.new_utxo(in_chain_id, in_amount, Some(index), None, None, rng)
+			.unwrap();
+		let in_utxo2 = prover
+			.new_utxo(in_chain_id, in_amount, Some(index), None, None, rng)
+			.unwrap();
 
 		// Adding invalid nullifier
-		in_utxos[0].nullifier = BnFr::rand(rng);
+		in_utxo1.nullifier = Some(BnFr::rand(rng));
 
-		let out_chain_ids = [0; 2];
-		let out_amounts = [5; 2];
-		let out_utxos = prover.new_utxos(out_chain_ids, out_amounts, rng).unwrap();
+		let in_utxos = [in_utxo1, in_utxo2];
 
-		let commitments = in_utxos.iter().map(|x| x.commitment).collect::<Vec<BnFr>>();
-		let (in_indices, in_paths, in_set_private_inputs, in_root_set) =
-			prover.setup_tree_and_set(&commitments).unwrap();
+		// Output Utxos
+		let out_chain_id = BnFr::from(0u32);
+		let out_amount = BnFr::from(10u32);
+		let out_utxo1 = prover
+			.new_utxo(out_chain_id, out_amount, None, None, None, rng)
+			.unwrap();
+		let out_utxo2 = prover
+			.new_utxo(out_chain_id, out_amount, None, None, None, rng)
+			.unwrap();
+		let out_utxos = [out_utxo1, out_utxo2];
 
-		let pub_ins = VAnchorProverBn2542x2::construct_public_inputs(
-			in_utxos[0].leaf_public.chain_id,
-			public_amount,
-			in_root_set.to_vec(),
-			in_utxos.iter().map(|x| x.nullifier).collect::<Vec<BnFr>>(),
-			out_utxos
-				.iter()
-				.map(|x| x.commitment)
-				.collect::<Vec<BnFr>>(),
-			ext_data_hash,
-		);
+		let leaf0 = in_utxos[0].commitment;
+		let (in_path0, _) = prover.setup_tree(&vec![leaf0], 0).unwrap();
+		let root0 = in_path0.root_hash(&leaf0).unwrap().inner();
+		let leaf1 = in_utxos[1].commitment;
+		let (in_path1, _) = prover.setup_tree(&vec![leaf1], 0).unwrap();
+		let root1 = in_path1.root_hash(&leaf1).unwrap().inner();
 
-		let circuit = prover
-			.setup_circuit(
+		let in_leaves = [vec![leaf0], vec![leaf1]];
+		let in_indices = [0; 2];
+		let in_root_set = [root0, root1];
+
+		let (circuit, .., pub_ins) = prover
+			.setup_circuit_with_utxos(
 				public_amount,
-				arbitrary_data,
-				in_utxos,
-				in_indices,
-				in_paths,
-				in_set_private_inputs,
+				ext_data_hash,
 				in_root_set,
+				in_indices,
+				in_leaves,
+				in_utxos,
 				out_utxos,
 			)
 			.unwrap();
@@ -627,44 +719,48 @@ mod test {
 
 		let public_amount = BnFr::from(0u32);
 		let ext_data_hash = BnFr::rand(rng);
-		let arbitrary_data = VAnchorProverBn2542x2::setup_arbitrary_data(ext_data_hash);
 
-		let in_chain_ids = [0; 2];
-		let in_amounts = [5; 2];
-		let mut in_utxos = prover.new_utxos(in_chain_ids, in_amounts, rng).unwrap();
+		// Input Utxos
+		let in_chain_id = BnFr::from(0u32);
+		let in_amount = BnFr::from(5u32);
+		let index = BnFr::from(0u32);
+		let in_utxo1 = prover
+			.new_utxo(in_chain_id, in_amount, Some(index), None, None, rng)
+			.unwrap();
 
-		// Enure both nullifiers are the same
-		in_utxos[0] = in_utxos[1].clone();
+		// Both inputs are the same -- attempt of double spending
+		let in_utxos = [in_utxo1, in_utxo1];
 
-		let out_chain_ids = [0; 2];
-		let out_amounts = [5; 2];
-		let out_utxos = prover.new_utxos(out_chain_ids, out_amounts, rng).unwrap();
+		// Output Utxos
+		let out_chain_id = BnFr::from(0u32);
+		let out_amount = BnFr::from(10u32);
+		let out_utxo1 = prover
+			.new_utxo(out_chain_id, out_amount, None, None, None, rng)
+			.unwrap();
+		let out_utxo2 = prover
+			.new_utxo(out_chain_id, out_amount, None, None, None, rng)
+			.unwrap();
+		let out_utxos = [out_utxo1, out_utxo2];
 
-		let commitments = in_utxos.iter().map(|x| x.commitment).collect::<Vec<BnFr>>();
-		let (in_indices, in_paths, in_set_private_inputs, in_root_set) =
-			prover.setup_tree_and_set(&commitments).unwrap();
+		let leaf0 = in_utxos[0].commitment;
+		let (in_path0, _) = prover.setup_tree(&vec![leaf0], 0).unwrap();
+		let root0 = in_path0.root_hash(&leaf0).unwrap().inner();
+		let leaf1 = in_utxos[1].commitment;
+		let (in_path1, _) = prover.setup_tree(&vec![leaf1], 0).unwrap();
+		let root1 = in_path1.root_hash(&leaf1).unwrap().inner();
 
-		let pub_ins = VAnchorProverBn2542x2::construct_public_inputs(
-			in_utxos[0].leaf_public.chain_id,
-			public_amount,
-			in_root_set.to_vec(),
-			in_utxos.iter().map(|x| x.nullifier).collect::<Vec<BnFr>>(),
-			out_utxos
-				.iter()
-				.map(|x| x.commitment)
-				.collect::<Vec<BnFr>>(),
-			ext_data_hash,
-		);
+		let in_leaves = [vec![leaf0], vec![leaf1]];
+		let in_indices = [0; 2];
+		let in_root_set = [root0, root1];
 
-		let circuit = prover
-			.setup_circuit(
+		let (circuit, .., pub_ins) = prover
+			.setup_circuit_with_utxos(
 				public_amount,
-				arbitrary_data,
-				in_utxos,
-				in_indices,
-				in_paths,
-				in_set_private_inputs,
+				ext_data_hash,
 				in_root_set,
+				in_indices,
+				in_leaves,
+				in_utxos,
 				out_utxos,
 			)
 			.unwrap();
@@ -688,41 +784,50 @@ mod test {
 
 		let public_amount = BnFr::from(10u32);
 		let ext_data_hash = BnFr::rand(rng);
-		let arbitrary_data = VAnchorProverBn2542x2::setup_arbitrary_data(ext_data_hash);
 
-		let in_chain_ids = [0; 2];
-		let in_amounts = [10; 2];
-		let in_utxos = prover.new_utxos(in_chain_ids, in_amounts, rng).unwrap();
+		// Input Utxos
+		let in_chain_id = BnFr::from(0u32);
+		// Input amount too high
+		let in_amount = BnFr::from(10u32);
+		let index = BnFr::from(0u32);
+		let in_utxo1 = prover
+			.new_utxo(in_chain_id, in_amount, Some(index), None, None, rng)
+			.unwrap();
+		let in_utxo2 = prover
+			.new_utxo(in_chain_id, in_amount, Some(index), None, None, rng)
+			.unwrap();
+		let in_utxos = [in_utxo1, in_utxo2];
 
-		let out_chain_ids = [0; 2];
-		let out_amounts = [5; 2];
-		let out_utxos = prover.new_utxos(out_chain_ids, out_amounts, rng).unwrap();
+		// Output Utxos
+		let out_chain_id = BnFr::from(0u32);
+		let out_amount = BnFr::from(10u32);
+		let out_utxo1 = prover
+			.new_utxo(out_chain_id, out_amount, None, None, None, rng)
+			.unwrap();
+		let out_utxo2 = prover
+			.new_utxo(out_chain_id, out_amount, None, None, None, rng)
+			.unwrap();
+		let out_utxos = [out_utxo1, out_utxo2];
 
-		let commitments = in_utxos.iter().map(|x| x.commitment).collect::<Vec<BnFr>>();
-		let (in_indices, in_paths, in_set_private_inputs, in_root_set) =
-			prover.setup_tree_and_set(&commitments).unwrap();
+		let leaf0 = in_utxos[0].commitment;
+		let (in_path0, _) = prover.setup_tree(&vec![leaf0], 0).unwrap();
+		let root0 = in_path0.root_hash(&leaf0).unwrap().inner();
+		let leaf1 = in_utxos[1].commitment;
+		let (in_path1, _) = prover.setup_tree(&vec![leaf1], 0).unwrap();
+		let root1 = in_path1.root_hash(&leaf1).unwrap().inner();
 
-		let pub_ins = VAnchorProverBn2542x2::construct_public_inputs(
-			in_utxos[0].leaf_public.chain_id,
-			public_amount,
-			in_root_set.to_vec(),
-			in_utxos.iter().map(|x| x.nullifier).collect::<Vec<BnFr>>(),
-			out_utxos
-				.iter()
-				.map(|x| x.commitment)
-				.collect::<Vec<BnFr>>(),
-			ext_data_hash,
-		);
+		let in_leaves = [vec![leaf0], vec![leaf1]];
+		let in_indices = [0; 2];
+		let in_root_set = [root0, root1];
 
-		let circuit = prover
-			.setup_circuit(
+		let (circuit, .., pub_ins) = prover
+			.setup_circuit_with_utxos(
 				public_amount,
-				arbitrary_data,
-				in_utxos,
-				in_indices,
-				in_paths,
-				in_set_private_inputs,
+				ext_data_hash,
 				in_root_set,
+				in_indices,
+				in_leaves,
+				in_utxos,
 				out_utxos,
 			)
 			.unwrap();
@@ -752,41 +857,49 @@ mod test {
 
 		let public_amount = BnFr::zero();
 		let ext_data_hash = BnFr::rand(rng);
-		let arbitrary_data = VAnchorProverBn2542x2::setup_arbitrary_data(ext_data_hash);
 
-		let in_chain_ids = [BnFr::from(0u32); 2];
-		let in_amounts = [limit + BnFr::one(); 2];
-		let in_utxos = prover.new_utxos_f(in_chain_ids, in_amounts, rng).unwrap();
+		// Input Utxos
+		let in_chain_id = BnFr::from(0u32);
+		let in_amount = BnFr::from(limit + BnFr::one());
+		let index = BnFr::from(0u32);
+		let in_utxo1 = prover
+			.new_utxo(in_chain_id, in_amount, Some(index), None, None, rng)
+			.unwrap();
+		let in_utxo2 = prover
+			.new_utxo(in_chain_id, in_amount, Some(index), None, None, rng)
+			.unwrap();
+		let in_utxos = [in_utxo1, in_utxo2];
 
-		let out_chain_ids = [BnFr::from(0u32); 2];
-		let out_amounts = [limit + BnFr::one(); 2];
-		let out_utxos = prover.new_utxos_f(out_chain_ids, out_amounts, rng).unwrap();
+		// Output Utxos
+		let out_chain_id = BnFr::from(0u32);
+		let out_amount = BnFr::from(10u32);
+		let out_utxo1 = prover
+			.new_utxo(out_chain_id, out_amount, None, None, None, rng)
+			.unwrap();
+		let out_utxo2 = prover
+			.new_utxo(out_chain_id, out_amount, None, None, None, rng)
+			.unwrap();
+		let out_utxos = [out_utxo1, out_utxo2];
 
-		let commitments = in_utxos.iter().map(|x| x.commitment).collect::<Vec<BnFr>>();
-		let (in_indices, in_paths, in_set_private_inputs, in_root_set) =
-			prover.setup_tree_and_set(&commitments).unwrap();
+		let leaf0 = in_utxos[0].commitment;
+		let (in_path0, _) = prover.setup_tree(&vec![leaf0], 0).unwrap();
+		let root0 = in_path0.root_hash(&leaf0).unwrap().inner();
+		let leaf1 = in_utxos[1].commitment;
+		let (in_path1, _) = prover.setup_tree(&vec![leaf1], 0).unwrap();
+		let root1 = in_path1.root_hash(&leaf1).unwrap().inner();
 
-		let pub_ins = VAnchorProverBn2542x2::construct_public_inputs(
-			in_utxos[0].leaf_public.chain_id,
-			public_amount,
-			in_root_set.to_vec(),
-			in_utxos.iter().map(|x| x.nullifier).collect::<Vec<BnFr>>(),
-			out_utxos
-				.iter()
-				.map(|x| x.commitment)
-				.collect::<Vec<BnFr>>(),
-			ext_data_hash,
-		);
+		let in_leaves = [vec![leaf0], vec![leaf1]];
+		let in_indices = [0; 2];
+		let in_root_set = [root0, root1];
 
-		let circuit = prover
-			.setup_circuit(
+		let (circuit, .., pub_ins) = prover
+			.setup_circuit_with_utxos(
 				public_amount,
-				arbitrary_data,
-				in_utxos,
-				in_indices,
-				in_paths,
-				in_set_private_inputs,
+				ext_data_hash,
 				in_root_set,
+				in_indices,
+				in_leaves,
+				in_utxos,
 				out_utxos,
 			)
 			.unwrap();
@@ -810,41 +923,49 @@ mod test {
 
 		let public_amount = BnFr::from(0u32);
 		let ext_data_hash = BnFr::rand(rng);
-		let arbitrary_data = VAnchorProverBn2542x2::setup_arbitrary_data(ext_data_hash);
 
-		let in_chain_ids = [0; 2];
-		let in_amounts = [5; 2];
-		let in_utxos = prover.new_utxos(in_chain_ids, in_amounts, rng).unwrap();
+		// Input Utxos
+		let in_chain_id = BnFr::from(0u32);
+		let in_amount = BnFr::from(5u32);
+		let index = BnFr::from(0u32);
+		let in_utxo1 = prover
+			.new_utxo(in_chain_id, in_amount, Some(index), None, None, rng)
+			.unwrap();
+		let in_utxo2 = prover
+			.new_utxo(in_chain_id, in_amount, Some(index), None, None, rng)
+			.unwrap();
+		let in_utxos = [in_utxo1, in_utxo2];
 
-		let out_chain_ids = [0; 2];
-		let out_amounts = [5; 2];
-		let out_utxos = prover.new_utxos(out_chain_ids, out_amounts, rng).unwrap();
+		// Output Utxos
+		let out_chain_id = BnFr::from(0u32);
+		let out_amount = BnFr::from(10u32);
+		let out_utxo1 = prover
+			.new_utxo(out_chain_id, out_amount, None, None, None, rng)
+			.unwrap();
+		let out_utxo2 = prover
+			.new_utxo(out_chain_id, out_amount, None, None, None, rng)
+			.unwrap();
+		let out_utxos = [out_utxo1, out_utxo2];
 
-		let commitments = in_utxos.iter().map(|x| x.commitment).collect::<Vec<BnFr>>();
-		let (in_indices, in_paths, in_set_private_inputs, in_root_set) =
-			prover.setup_tree_and_set(&commitments).unwrap();
+		let leaf0 = in_utxos[0].commitment;
+		let (in_path0, _) = prover.setup_tree(&vec![leaf0], 0).unwrap();
+		let root0 = in_path0.root_hash(&leaf0).unwrap().inner();
+		let leaf1 = in_utxos[1].commitment;
+		let (in_path1, _) = prover.setup_tree(&vec![leaf1], 0).unwrap();
+		let root1 = in_path1.root_hash(&leaf1).unwrap().inner();
 
-		let pub_ins = VAnchorProverBn2542x2::construct_public_inputs(
-			in_utxos[0].leaf_public.chain_id,
-			public_amount,
-			in_root_set.to_vec(),
-			in_utxos.iter().map(|x| x.nullifier).collect::<Vec<BnFr>>(),
-			out_utxos
-				.iter()
-				.map(|x| x.commitment)
-				.collect::<Vec<BnFr>>(),
-			ext_data_hash,
-		);
+		let in_leaves = [vec![leaf0], vec![leaf1]];
+		let in_indices = [0; 2];
+		let in_root_set = [root0, root1];
 
-		let circuit = prover
-			.setup_circuit(
+		let (circuit, .., pub_ins) = prover
+			.setup_circuit_with_utxos(
 				public_amount,
-				arbitrary_data,
-				in_utxos,
-				in_indices,
-				in_paths,
-				in_set_private_inputs,
+				ext_data_hash,
 				in_root_set,
+				in_indices,
+				in_leaves,
+				in_utxos,
 				out_utxos,
 			)
 			.unwrap();
