@@ -7,22 +7,18 @@ use crate::{
 };
 use ark_bn254::Fr as Bn254Fr;
 use ark_crypto_primitives::Error;
-use ark_ff::{BigInteger, PrimeField};
-use ark_std::{rand::RngCore, rc::Rc, vec::Vec};
+use ark_ff::PrimeField;
+use ark_std::{error::Error as ArkError, rand::RngCore, rc::Rc, string::ToString, vec::Vec};
 use arkworks_gadgets::{
 	arbitrary::vanchor_data::VAnchorArbitraryData,
 	keypair::vanchor::Keypair,
 	leaf::vanchor::{Private as LeafPrivateInput, Public as LeafPublicInput, VAnchorLeaf as Leaf},
 	merkle_tree::Path,
-	set::membership::{Private as SetPrivateInputs, SetMembership},
 };
 use arkworks_utils::{
 	poseidon::PoseidonParameters,
-	utils::{
-		common::{
-			setup_params_x5_2, setup_params_x5_3, setup_params_x5_4, setup_params_x5_5, Curve,
-		},
-		keccak_256, ExtData,
+	utils::common::{
+		setup_params_x5_2, setup_params_x5_3, setup_params_x5_4, setup_params_x5_5, Curve,
 	},
 };
 
@@ -42,17 +38,92 @@ pub fn get_hash_params<F: PrimeField>(
 	)
 }
 
-#[derive(Clone)]
+#[derive(Debug)]
+pub enum UtxoError {
+	NullifierNotCalculated,
+}
+
+impl core::fmt::Display for UtxoError {
+	fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
+		let msg = match self {
+			UtxoError::NullifierNotCalculated => "Nullifier not calculated".to_string(),
+		};
+		write!(f, "{}", msg)
+	}
+}
+
+impl ArkError for UtxoError {}
+
+#[derive(Clone, Copy)]
 pub struct Utxo<F: PrimeField> {
 	pub chain_id: F,
 	pub amount: F,
 	pub keypair: Keypair<F, PoseidonCRH_x5_2<F>>,
 	pub leaf_private: LeafPrivateInput<F>,
 	pub leaf_public: LeafPublicInput<F>,
-	pub nullifier: F,
+	pub index: Option<F>,
+	pub nullifier: Option<F>,
 	pub commitment: F,
 }
 
+impl<F: PrimeField> Utxo<F> {
+	pub fn new<R: RngCore>(
+		chain_id: F,
+		amount: F,
+		index: Option<F>,
+		private_key: Option<F>,
+		blinding: Option<F>,
+		params2: &PoseidonParameters<F>,
+		params4: &PoseidonParameters<F>,
+		params5: &PoseidonParameters<F>,
+		rng: &mut R,
+	) -> Result<Self, Error> {
+		let blinding = blinding.unwrap_or(F::rand(rng));
+		let private_input = LeafPrivateInput::<F>::new(amount, blinding);
+		let public_input = LeafPublicInput::<F>::new(chain_id);
+
+		let keypair = Keypair::new(private_key.unwrap_or(F::rand(rng)));
+		let pub_key = keypair.public_key(params2)?;
+
+		let leaf = Leaf::<F, PoseidonCRH_x5_4<F>>::create_leaf(
+			&private_input,
+			&public_input,
+			&pub_key,
+			&params5,
+		)?;
+
+		let nullifier = if index.is_some() {
+			let i = index.unwrap();
+
+			let signature = keypair.signature(&leaf, &i, params4)?;
+
+			let nullifier =
+				Leaf::<_, PoseidonCRH_x5_4<F>>::create_nullifier(&signature, &leaf, &params4, &i)?;
+
+			Some(nullifier)
+		} else {
+			None
+		};
+
+		Ok(Self {
+			chain_id,
+			amount,
+			keypair,
+			leaf_private: private_input,
+			leaf_public: public_input,
+			index,
+			nullifier,
+			commitment: leaf,
+		})
+	}
+
+	pub fn get_nullifier(&self) -> Result<F, Error> {
+		self.nullifier
+			.ok_or(UtxoError::NullifierNotCalculated.into())
+	}
+}
+
+#[derive(Clone)]
 pub struct VAnchorProverSetup<
 	F: PrimeField,
 	const TREE_DEPTH: usize,
@@ -88,57 +159,26 @@ impl<
 		}
 	}
 
-	pub fn new_utxos<R: RngCore, const N: usize>(
+	pub fn new_utxo<R: RngCore>(
 		&self,
-		chain_ids: [u128; N],
-		amounts: [u128; N],
+		chain_id: F,
+		amount: F,
+		index: Option<F>,
+		secret_key: Option<F>,
+		blinding: Option<F>,
 		rng: &mut R,
-	) -> Result<Vec<Utxo<F>>, Error> {
-		let chain_ids_f = chain_ids.map(|x| F::from(x));
-		let amounts_f = amounts.map(|x| F::from(x));
-
-		let keypairs = Self::setup_keypairs::<_, N>(rng);
-		let (commitments, nullifiers, leaf_privates, leaf_publics) =
-			self.setup_leaves(&chain_ids_f, &amounts_f, &keypairs, rng)?;
-
-		let utxos: Vec<Utxo<F>> = (0..N)
-			.map(|i| Utxo {
-				chain_id: chain_ids_f[i],
-				amount: amounts_f[i],
-				keypair: keypairs[i].clone(),
-				leaf_private: leaf_privates[i].clone(),
-				leaf_public: leaf_publics[i].clone(),
-				nullifier: nullifiers[i],
-				commitment: commitments[i],
-			})
-			.collect();
-
-		Ok(utxos)
-	}
-
-	pub fn new_utxos_f<R: RngCore, const N: usize>(
-		&self,
-		chain_ids_f: [F; N],
-		amounts_f: [F; N],
-		rng: &mut R,
-	) -> Result<Vec<Utxo<F>>, Error> {
-		let keypairs = Self::setup_keypairs::<_, N>(rng);
-		let (commitments, nullifiers, leaf_privates, leaf_publics) =
-			self.setup_leaves(&chain_ids_f, &amounts_f, &keypairs, rng)?;
-
-		let utxos: Vec<Utxo<F>> = (0..N)
-			.map(|i| Utxo {
-				chain_id: chain_ids_f[i],
-				amount: amounts_f[i],
-				keypair: keypairs[i].clone(),
-				leaf_private: leaf_privates[i].clone(),
-				leaf_public: leaf_publics[i].clone(),
-				nullifier: nullifiers[i],
-				commitment: commitments[i],
-			})
-			.collect();
-
-		Ok(utxos)
+	) -> Result<Utxo<F>, Error> {
+		Utxo::new(
+			chain_id,
+			amount,
+			index,
+			secret_key,
+			blinding,
+			&self.params2,
+			&self.params4,
+			&self.params5,
+			rng,
+		)
 	}
 
 	pub fn setup_random_circuit<R: RngCore>(
@@ -159,34 +199,46 @@ impl<
 		>,
 		Error,
 	> {
-		let public_amount = rng.next_u64() as i128;
+		let public_amount = F::rand(rng);
+		let ext_data_hash = F::rand(rng);
+		let in_root_set = [F::rand(rng); M];
+		let in_leaves = [F::rand(rng); INS].map(|x| vec![x]);
+		let in_indices = [0; INS];
 
-		let mut recipient = [0u8; 20];
-		rng.fill_bytes(&mut recipient);
+		let chain_id = F::rand(rng);
+		let amount = F::rand(rng);
+		let index = F::rand(rng);
+		let secret_key = F::rand(rng);
+		let blinding = F::rand(rng);
 
-		let mut relayer = [0u8; 20];
-		rng.fill_bytes(&mut relayer);
-
-		let ext_amount = rng.next_u64() as i128;
-		let fee = rng.next_u64() as u128;
-
-		let in_chain_id = rng.next_u64() as u128;
-		let in_amounts = [rng.next_u64() as u128; INS];
-		let out_chain_ids = [rng.next_u64() as u128; OUTS];
-		let out_amounts = [rng.next_u64() as u128; OUTS];
-
-		let (circuit, ..) = self.setup_circuit_with_data(
-			public_amount,
-			recipient.to_vec(),
-			relayer.to_vec(),
-			ext_amount,
-			fee,
-			in_chain_id,
-			in_amounts,
-			out_chain_ids,
-			out_amounts,
+		let in_utxo = self.new_utxo(
+			chain_id,
+			amount,
+			Some(index),
+			Some(secret_key),
+			Some(blinding),
 			rng,
 		)?;
+		let in_utxos = [in_utxo; INS];
+		let out_utxo = self.new_utxo(
+			chain_id,
+			amount,
+			None,
+			Some(secret_key),
+			Some(blinding),
+			rng,
+		)?;
+		let out_utxos = [out_utxo; OUTS];
+
+		let (circuit, ..) = self.setup_circuit_with_utxos(
+			public_amount,
+			ext_data_hash,
+			in_root_set,
+			in_indices,
+			in_leaves,
+			in_utxos,
+			out_utxos,
+		)?;
 
 		Ok(circuit)
 	}
@@ -194,15 +246,15 @@ impl<
 	pub fn setup_circuit_with_utxos(
 		self,
 		// External data
-		public_amount: i128,
-		recipient: Vec<u8>,
-		relayer: Vec<u8>,
-		ext_amount: i128,
-		fee: u128,
+		public_amount: F,
+		ext_data_hash: F,
+		in_root_set: [F; M],
+		in_indices: [u64; INS],
+		in_leaves: [Vec<F>; INS],
 		// Input transactions
-		in_utxos: Vec<Utxo<F>>,
+		in_utxos: [Utxo<F>; INS],
 		// Output transactions
-		out_utxos: Vec<Utxo<F>>,
+		out_utxos: [Utxo<F>; OUTS],
 	) -> Result<
 		(
 			VACircuit<
@@ -221,124 +273,53 @@ impl<
 		),
 		Error,
 	> {
-		assert_eq!(in_utxos.len(), INS);
-		assert_eq!(out_utxos.len(), OUTS);
 		// Tree + set for proving input txos
-		let in_commitments = in_utxos.iter().map(|x| x.commitment).collect::<Vec<F>>();
-		let (in_indices, in_paths, in_set_private_inputs, in_root_set) =
-			self.setup_tree_and_set(&in_commitments[..])?;
-
-		let ext_data = ExtData::new(
-			recipient,
-			relayer,
-			ext_amount.to_le_bytes().to_vec(),
-			fee.to_le_bytes().to_vec(),
-			out_utxos[0].commitment.into_repr().to_bytes_le(),
-			out_utxos[1].commitment.into_repr().to_bytes_le(),
-		);
-		let ext_data_hash = keccak_256(&ext_data.encode_abi());
-		let ext_data_hash_f = F::from_le_bytes_mod_order(&ext_data_hash);
-		// Arbitrary data
-		let arbitrary_data = Self::setup_arbitrary_data(ext_data_hash_f);
-
-		let mut public_amount_f = F::from(public_amount.unsigned_abs());
-		if public_amount.is_negative() {
-			public_amount_f = -public_amount_f;
+		let in_indices_f = in_indices.map(|x| F::from(x));
+		let mut in_paths = Vec::new();
+		for i in 0..INS {
+			let (in_path, _) = self.setup_tree(&in_leaves[i], in_indices[i])?;
+			in_paths.push(in_path)
 		}
+		// Arbitrary data
+		let arbitrary_data = Self::setup_arbitrary_data(ext_data_hash);
 
 		let circuit = self.setup_circuit(
-			public_amount_f,
+			public_amount,
 			arbitrary_data,
 			in_utxos.clone(),
-			in_indices,
+			in_indices_f,
 			in_paths,
-			in_set_private_inputs,
 			in_root_set,
 			out_utxos.clone(),
 		)?;
 
-		let in_nullifiers = in_utxos.iter().map(|x| x.nullifier).collect::<Vec<F>>();
+		let in_nullifiers: Result<Vec<F>, Error> =
+			in_utxos.iter().map(|x| x.get_nullifier()).collect();
 		let out_nullifiers = out_utxos.iter().map(|x| x.commitment).collect::<Vec<F>>();
 		let public_inputs = Self::construct_public_inputs(
 			in_utxos[0].leaf_public.chain_id,
-			public_amount_f,
+			public_amount,
 			in_root_set.to_vec(),
-			in_nullifiers,
+			in_nullifiers?,
 			out_nullifiers,
-			ext_data_hash_f,
+			ext_data_hash,
 		);
 
 		Ok((circuit, public_inputs))
 	}
 
-	// This function is used only for first transaction, when the tree is empty
-	pub fn setup_circuit_with_data<R: RngCore>(
-		self,
-		public_amount: i128,
-		recipient: Vec<u8>,
-		relayer: Vec<u8>,
-		ext_amount: i128,
-		fee: u128,
-		in_chain_id: u128,
-		in_amounts: [u128; INS],
-		out_chain_ids: [u128; OUTS],
-		out_amounts: [u128; OUTS],
-		rng: &mut R,
-	) -> Result<
-		(
-			VACircuit<
-				F,
-				PoseidonCRH_x5_2<F>,
-				PoseidonCRH_x5_2Gadget<F>,
-				TreeConfig_x5<F>,
-				LeafCRHGadget<F>,
-				PoseidonCRH_x5_3Gadget<F>,
-				TREE_DEPTH,
-				INS,
-				OUTS,
-				M,
-			>,
-			Vec<F>,
-			Vec<Utxo<F>>,
-			Vec<Utxo<F>>,
-		),
-		Error,
-	> {
-		// Making a vec of same chain ids to be passed into setup_leaves
-		let in_chain_ids = [in_chain_id; INS];
-
-		// Input leaves (txos)
-		let in_utxos = self.new_utxos(in_chain_ids, in_amounts, rng)?;
-
-		// Output leaves (txos)
-		let out_utxos = self.new_utxos(out_chain_ids, out_amounts, rng)?;
-
-		let (circuit, public_inputs) = self.setup_circuit_with_utxos(
-			public_amount,
-			recipient,
-			relayer,
-			ext_amount,
-			fee,
-			in_utxos.clone(),
-			out_utxos.clone(),
-		)?;
-
-		Ok((circuit, public_inputs, in_utxos, out_utxos))
-	}
-
 	pub fn setup_circuit(
 		self,
 		public_amount: F,
 		arbitrary_data: VAnchorArbitraryData<F>,
 		// Input transactions
-		in_utxos: Vec<Utxo<F>>,
+		in_utxos: [Utxo<F>; INS],
 		// Data related to tree
-		in_indicies: Vec<F>,
+		in_indicies: [F; INS],
 		in_paths: Vec<Path<TreeConfig_x5<F>, TREE_DEPTH>>,
-		in_set_private_inputs: Vec<SetPrivateInputs<F, M>>,
 		in_root_set: [F; M],
 		// Output transactions
-		out_utxos: Vec<Utxo<F>>,
+		out_utxos: [Utxo<F>; OUTS],
 	) -> Result<
 		VACircuit<
 			F,
@@ -362,7 +343,8 @@ impl<
 			.iter()
 			.map(|x| x.keypair.clone())
 			.collect::<Vec<Keypair<F, PoseidonCRH_x5_2<F>>>>();
-		let in_nullifiers = in_utxos.iter().map(|x| x.nullifier).collect::<Vec<F>>();
+		let in_nullifiers: Result<Vec<F>, Error> =
+			in_utxos.iter().map(|x| x.get_nullifier()).collect();
 
 		let out_pub_keys: Result<Vec<F>, _> = out_utxos
 			.iter()
@@ -394,14 +376,13 @@ impl<
 			in_leaf_private_inputs,
 			in_keypair_inputs,
 			in_utxos[0].leaf_public.clone(),
-			in_set_private_inputs,
 			in_root_set,
 			self.params2,
 			self.params4,
 			self.params5,
 			in_paths,
-			in_indicies,
-			in_nullifiers,
+			in_indicies.to_vec(),
+			in_nullifiers?,
 			out_commitments,
 			out_leaf_private,
 			out_leaf_public,
@@ -417,113 +398,17 @@ impl<
 		[(); N].map(|_| Keypair::<_, PoseidonCRH_x5_2<F>>::new(F::rand(rng)))
 	}
 
-	pub fn setup_leaves<R: RngCore, const N: usize>(
-		&self,
-		chain_ids: &[F; N],
-		amounts: &[F; N],
-		keypairs: &[Keypair<F, PoseidonCRH_x5_2<F>>; N],
-		rng: &mut R,
-	) -> Result<
-		(
-			[F; N],
-			[F; N],
-			[LeafPrivateInput<F>; N],
-			[LeafPublicInput<F>; N],
-		),
-		Error,
-	> {
-		let num_inputs = amounts.len();
-
-		let mut leaves = [F::default(); N];
-		let mut nullifiers = [F::default(); N];
-		let mut private_inputs = [LeafPrivateInput::<F>::default(); N];
-		let mut public_inputs = [LeafPublicInput::<F>::default(); N];
-
-		for i in 0..num_inputs {
-			let chain_id = F::from(chain_ids[i]);
-			let amount = F::from(amounts[i]);
-			let blinding = F::rand(rng);
-			let index = F::from(i as u64);
-
-			let private_input = LeafPrivateInput::<F>::new(amount, blinding);
-			let public_input = LeafPublicInput::<F>::new(chain_id);
-
-			let pub_key = keypairs[i].public_key(&self.params2)?;
-
-			let leaf = Leaf::<F, PoseidonCRH_x5_4<F>>::create_leaf(
-				&private_input,
-				&public_input,
-				&pub_key,
-				&self.params5,
-			)?;
-
-			let signature = keypairs[i].signature(&leaf, &index, &self.params4)?;
-
-			let nullifier = Leaf::<F, PoseidonCRH_x5_4<F>>::create_nullifier(
-				&signature,
-				&leaf,
-				&self.params4,
-				&index,
-			)?;
-
-			leaves[i] = leaf;
-			nullifiers[i] = nullifier;
-			private_inputs[i] = private_input;
-			public_inputs[i] = public_input;
-		}
-
-		Ok((leaves, nullifiers, private_inputs, public_inputs))
-	}
-
 	pub fn setup_tree(
 		&self,
 		leaves: &[F],
-	) -> Result<(Vec<Path<TreeConfig_x5<F>, TREE_DEPTH>>, Vec<F>, F), Error> {
+		index: u64,
+	) -> Result<(Path<TreeConfig_x5<F>, TREE_DEPTH>, F), Error> {
 		let inner_params = Rc::new(self.params3.clone());
 		let tree = Tree_x5::new_sequential(inner_params, Rc::new(()), &leaves.to_vec())?;
 		let root = tree.root();
+		let path = tree.generate_membership_proof::<TREE_DEPTH>(index);
 
-		let num_leaves = leaves.len();
-
-		let mut paths = Vec::new();
-		let mut indices = Vec::new();
-		for i in 0..num_leaves {
-			let path = tree.generate_membership_proof::<TREE_DEPTH>(i as u64);
-			let index = path.get_index(&root, &leaves[i])?;
-			paths.push(path);
-			indices.push(index);
-		}
-
-		Ok((paths, indices, root.inner()))
-	}
-
-	pub fn setup_root_set(root: F) -> Result<([F; M], Vec<SetPrivateInputs<F, M>>), Error> {
-		let root_set = [root.clone(); M];
-
-		let mut set_private_inputs = Vec::new();
-		for _ in 0..M {
-			let set_private_input = SetMembership::generate_secrets(&root, &root_set)?;
-			set_private_inputs.push(set_private_input);
-		}
-
-		Ok((root_set, set_private_inputs))
-	}
-
-	pub fn setup_tree_and_set(
-		&self,
-		leaves: &[F],
-	) -> Result<
-		(
-			Vec<F>,
-			Vec<Path<TreeConfig_x5<F>, TREE_DEPTH>>,
-			Vec<SetPrivateInputs<F, M>>,
-			[F; M],
-		),
-		Error,
-	> {
-		let (paths, indices, root) = self.setup_tree(&leaves)?;
-		let (root_set, set_private_inputs) = Self::setup_root_set(root)?;
-		Ok((indices, paths, set_private_inputs, root_set))
+		Ok((path, root.inner()))
 	}
 
 	pub fn construct_public_inputs(
@@ -543,7 +428,7 @@ impl<
 		public_inputs
 	}
 
-	// TODO: Fix to match INS and OUTS
+	// NOTE: To be used for testing
 	pub fn deconstruct_public_inputs(
 		public_inputs: Vec<F>,
 	) -> Result<
@@ -559,27 +444,25 @@ impl<
 	> {
 		let mut pub_ins = public_inputs;
 
-		let err = "Invalid public inputs";
-
 		let mut root_set = Vec::new();
 		for _ in 0..INS {
-			root_set.push(pub_ins.pop().ok_or(err)?);
+			root_set.push(pub_ins.pop().unwrap());
 		}
 
-		let chain_id = pub_ins.pop().ok_or(err)?;
+		let chain_id = pub_ins.pop().unwrap();
 
 		let mut out_commitments = Vec::new();
 		for _ in 0..OUTS {
-			out_commitments.push(pub_ins.pop().ok_or(err)?);
+			out_commitments.push(pub_ins.pop().unwrap());
 		}
 
 		let mut in_nullifiers = Vec::new();
 		for _ in 0..INS {
-			in_nullifiers.push(pub_ins.pop().ok_or(err)?);
+			in_nullifiers.push(pub_ins.pop().unwrap());
 		}
 
-		let ext_data_hash = pub_ins.pop().ok_or(err)?;
-		let public_amount = pub_ins.pop().ok_or(err)?;
+		let ext_data_hash = pub_ins.pop().unwrap();
+		let public_amount = pub_ins.pop().unwrap();
 
 		Ok((
 			chain_id,
